from flask import Flask, request, abort, send_from_directory
from flask_migrate import Migrate
from flask_limiter import Limiter
from flask_cors import CORS
from flask_jwt import JWT, jwt_required, JWTError
from werkzeug.security import safe_str_cmp
from youtube_dl import YoutubeDL
import flask_limiter.util
import os
import json
import uuid
import datetime

from models import db, Sample
from celery_context_patch import make_celery


app = Flask(__name__)
app.config.from_json('config.json')

# Setup DB
db.app = app
db.init_app(app)
migrate = Migrate(app, db)
# Setup limiter
limiter = Limiter(app, key_func=flask_limiter.util.get_ipaddr, headers_enabled=True)
# Setup CORS
CORS(app)
# Setup celery
celery = make_celery(app)
# Setup JWT
jwt = JWT()
app.config['JWT_EXPIRATION_DELTA'] = datetime.timedelta(seconds=3000)
app.config['JWT_AUTH_URL_RULE'] = '/moderation/auth'


class SingletonUser:
    def __init__(self):
        self.id = 'the only one'


@jwt.authentication_handler
def check_auth(secret):
    if safe_str_cmp(secret.encode('utf-8'), app.config['SECRET_KEY'].encode('utf-8')):
        return SingletonUser()


@jwt.auth_request_handler
def authenticate():
    data = request.get_json()
    secret = data.get('secret') if data is not None else None
    if secret is None:
        raise JWTError('Bad Request', 'Malformed request structure')

    identity = jwt.authentication_callback(secret)
    if identity:
        token = jwt.jwt_encode_callback(identity)
        return jwt.auth_response_callback(token, identity)

    raise JWTError('Bad Request', 'Invalid credentials')


@jwt.identity_handler
def get_user(payload):
    return SingletonUser()


jwt.init_app(app)


# Setup over - lets get down to business
@app.route('/submit', methods=['POST'])
def upload():
    if not ('file' in request.files and request.form['submitter']):
        abort(400)

    file_uuid = str(uuid.uuid4())
    file = request.files['file']
    title = file.filename
    submitter = request.form['submitter']

    fullpath = os.path.join(app.config['SAMPLE_DIR'], file_uuid)

    sample = Sample(file_uuid, title, fullpath, submitter)
    db.session.add(sample)
    db.session.commit()
    file.save(fullpath)

    return json.dumps({'status': 'success', 'id': sample.uuid}), 201


@app.route('/submit/url', methods=['POST'])
def from_url():
    data = request.get_json()
    if not (data and data.get('url') and data.get('submitter')):
        abort(400)

    url = data['url']
    submitter = data['submitter']
    sample_uuid = str(uuid.uuid4())

    fullpath = os.path.join(app.config['SAMPLE_DIR'], f'{sample_uuid}.mp3')

    sample = Sample(sample_uuid, url, fullpath, submitter, transient=True, progress=-1)
    db.session.add(sample)
    db.session.commit()
    download_video.delay(url, sample_uuid)

    return json.dumps({'status': 'success', 'id': sample_uuid}), 201


@app.route('/status', methods=['POST'])
def check_status():
    data = request.get_json()
    if data is None or data.get('samples') is None or type(data['samples']) != list or len(data['samples']) == 0:
        abort(400)

    sampledata = []
    for sample in Sample.query.filter(Sample.uuid.in_(data['samples'])):
        sampledata.append({
            'uuid': sample.uuid,
            'title': sample.name,
            'transient': sample.transient,
            'progress': sample.progress,
            'approved': sample.approved,
            'rejectionMessage': sample.rejectionMessage,
        })

    return json.dumps({'status': 'success', 'data': sampledata})


@app.route('/moderation/uploads', methods=['GET'])
@app.route('/moderation/uploads/<sampleid>', methods=['POST'])
@jwt_required()
def list_unaproved(sampleid=None):
    if request.method == 'GET':
        unaproved = Sample.query.filter_by(approved=False, rejectionMessage=None, transient=False).all()
        samples = [{'id': sample.uuid, 'name': sample.name, 'path': sample.path, 'uploader': sample.uploader} for sample in unaproved]
        return {'status': 'success', 'samples': samples}, 200
    else:
        sample = Sample.query.get(sampleid)
        if sample is None:
            abort(404)
        data = request.get_json()
        if data is None:
            abort(400)

        if data['approve']:
            sample.approved = True
        else:
            sample.rejectionMessage = data['reason']

        db.session.commit()
        return {'status': 'success'}, 200


@app.route('/moderation/vetted', methods=['GET'])
@jwt_required()
def list_aproved(sampleid=None):
    if request.method == 'GET':
        aproved = Sample.query.filter_by(approved=True, transient=False).all()
        samples = [{'id': sample.uuid, 'name': sample.name, 'path': sample.path, 'uploader': sample.uploader} for sample in aproved]
        return {'status': 'success', 'samples': samples}, 200


@app.route('/samples/<path:path>')
def files(path):
    return send_from_directory(app.config['SAMPLE_DIR'], path)


@celery.task
def download_video(url, sample_uuid):
    sample = Sample.query.get(sample_uuid)
    sample.progress = 0
    db.session.commit()

    def update_progress(data):
        if (data.get('total_bytes') is None or data.get('downloaded_bytes') is None):
            return

        if data['status'] == 'downloading':
            progress = data['downloaded_bytes'] / data['total_bytes']
        elif data['status'] == 'finished':
            progress = 1
        else:
            return

        sample = Sample.query.get(sample_uuid)
        sample.progress = progress
        db.session.commit()

    try:
        ytdl = YoutubeDL({
            'quiet': True,
            'ratelimit': 500_000,
            'format': 'bestaudio/best',
            'outtmpl': os.path.join(app.config['SAMPLE_DIR'], '%(custom_uuid)s.%(ext)s'),
            'postprocessors': [{'key': 'FFmpegExtractAudio', 'preferredcodec': 'mp3'}],
            'progress_hooks': [update_progress]
        })

        info = ytdl.extract_info(url, extra_info={'custom_uuid': sample_uuid})
        title = info['title']
    except Exception:
        sample = Sample.query.get(sample_uuid)
        sample.transient = False
        sample.rejectionMessage = 'Failed to download'
        db.session.commit()
        return

    sample = Sample.query.get(sample_uuid)
    sample.name = title
    sample.transient = False
    sample.progress = 2
    db.session.commit()


if __name__ == "__main__":
    app.run()
